# -*- coding: utf-8 -*-

import copy

from hlogedu.search.containers import HashableList
from hlogedu.search.problem import action, DDRange, Heuristic, Problem


class BlocksWorldProblem(Problem):

    NAME = "Blocks"

    def __init__(self):
        super().__init__()
        self.initial_state = [[6], [1, 2], [3, 5], [4]]
        self.final_state = [[1, 2, 3], [], [], [4, 5, 6]]
        self.max_blocks = 3
        self.n_positions = len(self.initial_state)

    def is_valid_state(self, state):
        return True

    def get_start_states(self):
        return [HashableList(self.initial_state)]

    def is_goal_state(self, state):
        return state == self.final_state

    @action(DDRange(0, "n_positions"), DDRange(0, "n_positions"), cost=1)
    def move(self, state, pos_s, pos_d):
        if pos_s == pos_d or not state[pos_s]:  # No move or no block
            return None
        if len(state[pos_d]) >= self.max_blocks:  # Destination full
            return None

        n_state = copy.deepcopy(state)
        n_state[pos_d].append(n_state[pos_s].pop())
        return n_state


@BlocksWorldProblem.heuristic
class Simple(Heuristic):
    """This heuristic only counts the number of places with
    incorrect blocks on top of it.
    """

    def compute(self, state):
        h = 0
        for goal_pos, s_pos in zip(self.problem.final_state, state):
            if goal_pos and not s_pos or not goal_pos and s_pos:
                # goal/pos has block and pos/goal hasn't
                h += 1
            elif goal_pos and s_pos:
                # block on top is different
                h += 1 if goal_pos[0] != s_pos[0] else 0

        return h
