# -*- coding: utf-8 -*-

import copy

from hlogedu.search.containers import HashableList
from hlogedu.search.problem import action, DRange, DDRange, Problem


class Hanoi1(Problem):

    def __init__(self):
        super().__init__()
        self.n_discs = 7
        self.n_rods = 3

    def get_start_states(self):
        return [
            HashableList(list(range(self.n_discs)) if r == 0 else
                         [] for r in range(self.n_rods))
        ]

    def is_goal_state(self, state):
        return len(state[-1]) == self.n_discs

    def is_valid_state(self, state):
        return True

    @action(DDRange(0, 'n_rods'), DDRange(0, 'n_rods'), cost=1)
    def move(self, state, rod_s, rod_d):
        if rod_s == rod_d or not state[rod_s]:
            return None
        if state[rod_d]:  # Rod 'd' has disks ?
            if state[rod_s][-1] < state[rod_d][-1]:
                return None

        new_state = copy.deepcopy(state)
        new_state[rod_d].append(new_state[rod_s].pop())

        return new_state


class Hanoi2(Problem):

    def __init__(self):
        super().__init__()
        self.n_discs = 7
        self.n_rods = 3

    def get_start_states(self):
        return [tuple(0 for _ in range(self.n_discs))]

    def is_goal_state(self, state):
        return all(x == self.n_rods - 1 for x in state)

    def is_valid_state(self, state):
        return True

    @action(DRange(0, 3), DRange(0, 3), cost=1)
    def move(self, state, rod_s, rod_d):
        if rod_s == rod_d:
            return None

        s_top = self.get_rod_top_disk(state, rod_s)
        d_top = self.get_rod_top_disk(state, rod_d)
        if s_top == d_top or s_top is None:
            return None
        if d_top is not None:
            if s_top > d_top:
                return None

        new_state = list(state)
        new_state[s_top] = rod_d
        return tuple(new_state)

    def get_rod_top_disk(self, state, rod):
        try:
            return state.index(rod)
        except ValueError:
            return None
